class Config:
    defaults = {
        "user_input_valid_color": 3,
        "user_input_invalid_color": 2,
        "reference_text_color": 8,
        "stats_template": "\n\nwpm: {stats.wpm:0.2f}\ntime: {stats.total_seconds:0.2f}s",
        "stats_color": 5,
        "summary_template": (
            "WPM: {stats.wpm:0.2f}\n"
            "CPM: {stats.cpm:0.2f}\n"
            "RAW WPM: {stats.raw_wpm:0.2f}\n"
            "RAW CPM: {stats.raw_cpm:0.2f}\n"
            "total seconds: {stats.total_seconds:0.2f}\n"
            "total minutes: {stats.total_minutes:0.2f}\n"
            "correct words: {stats.correct_words}\n"
            "correct chars: {stats.correct_chars}\n"
            "incorrect words: {stats.incorrect_words}\n"
            "incorrect chars: {stats.incorrect_chars}\n"
            "total words: {stats.total_words}\n"
            "total chars: {stats.total_chars}\n"
            "accuracy: {stats.accuracy:0.2f}%"
        ),
        "summary_datafile": "~/.cache/fasttyper/datafile.csv",
        "top_margin_percentage": 30,
        "left_margin_percentage": 10,
        "lines_on_screen": 3,
    }

    def __init__(self, configmap):
        self.configmap = configmap

    def get(self, key):
        return self.configmap.get(key, self.defaults[key])



import curses


class Interface:
    def __init__(self, application, components, no_cursor=False):
        self.application = application
        self.components = components
        self.no_cursor = no_cursor

    def init_colors(self):
        assert curses.has_colors()
        curses.start_color()
        curses.use_default_colors()

        for i in range(0, curses.COLORS):
            curses.init_pair(i + 1, i, -1)

    def init(self, screen):
        self.init_colors()

    def update(self, screen):
        screen.clear()
        for component in self.components:
            component.paint(screen, self.application)
        screen.refresh()

    def __call__(self, screen):
        """
        Main running loop
        """
        self.init(screen)
        self.application.start()

        if self.no_cursor:
            curses.curs_set(0)

        while self.application.running():
            self.update(screen)
            self.application.action(screen)



from .listener import Action
from datetime import datetime
import os
import csv
import pathlib


class Stats:
    def __init__(self):
        self.correct_words = 0
        self.correct_chars = 0
        self.incorrect_words = 0
        self.incorrect_chars = 0

        self.start_dtime = None
        self.stop_dtime = None

    def signal_running(self):
        if self.start_dtime is None:
            self.start_dtime = datetime.now()

    def update(self, action, valid, running):
        if running:
            self.signal_running()

        if action == Action.add_char and valid is True:
            self.correct_chars += 1
        elif action == Action.add_char and valid is False:
            self.incorrect_chars += 1
        elif action in (Action.add_space, Action.add_newline) and valid is True:
            self.correct_chars += 1
            self.correct_words += 1
        elif action in (Action.add_space, Action.add_newline) and valid is False:
            self.incorrect_chars += 1
            self.incorrect_words += 1

        if not running:
            self.signal_stop()

            if valid:
                self.correct_words += 1
            else:
                self.incorrect_words += 1

    def signal_stop(self):
        self.stop_dtime = datetime.now()

    @property
    def total_seconds(self):
        stop_dtime = self.stop_dtime or datetime.now()
        start_dtime = self.start_dtime or datetime.now()
        return (stop_dtime - start_dtime).total_seconds() or 1

    @property
    def total_minutes(self):
        return self.total_seconds / 60

    @property
    def total_words(self):
        return self.incorrect_words + self.correct_words

    @property
    def total_chars(self):
        return self.incorrect_chars + self.correct_chars

    @property
    def wpm(self):
        return self.cpm / 5

    @property
    def cpm(self):
        return self.correct_chars / self.total_minutes

    @property
    def raw_wpm(self):
        return self.raw_cpm / 5

    @property
    def raw_cpm(self):
        return self.total_chars / self.total_minutes

    @property
    def accuracy(self):
        if self.total_chars:
            return self.correct_chars / self.total_chars * 100
        return 100

    def summarize(self, template):
        print(template.format(stats=self))

    def produce_record(self):
        return {
            "start_dtime": self.start_dtime.isoformat(),
            "stop_dtime": self.stop_dtime.isoformat(),
            "total_seconds": self.total_seconds,
            "total_minutes": self.total_minutes,
            "total_chars": self.total_chars,
            "correct_chars": self.correct_chars,
            "incorrect_chars": self.incorrect_chars,
            "total_words": self.total_words,
            "correct_words": self.correct_words,
            "incorrect_words": self.incorrect_words,
            "wpm": self.wpm,
            "cpm": self.cpm,
            "raw_wpm": self.raw_wpm,
            "raw_cpm": self.raw_cpm,
            "accuracy": self.accuracy,
        }

    def export_to_datafile(self, datafile):
        if datafile is None:
            return

        datafile = os.path.expanduser(datafile)
        data_dir, filename = os.path.split(datafile)

        pathlib.Path(data_dir).mkdir(exist_ok=True, parents=True)
        exists = os.path.isfile(datafile)

        record = self.produce_record()

        if not exists:
            with open(datafile, "w") as f:
                writter = csv.DictWriter(f, fieldnames=record.keys())
                writter.writeheader()
                writter.writerow(record)
        else:
            with open(datafile, "a") as f:
                writter = csv.DictWriter(f, fieldnames=record.keys())
                writter.writerow(record)

        print(f"\nwrote stats to {datafile}")



import sys


class Application:
    def __init__(self, listener, user_buffer, reference_buffer, config):
        self.listener = listener
        self.user_buffer = user_buffer
        self.reference_buffer = reference_buffer
        self.config = config
        self.finished = False
        self.silent_exit = False

        from .state import StateMachine

        self.state = StateMachine(self)

        from .stats import Stats

        self.stats = Stats()

    def start(self):
        self.state.signal_start()

    def running(self):
        return self.state.running()

    def valid_user_text_position(self):
        if self.state.mistake_position is not None:
            return self.state.mistake_position
        return self.user_buffer.get_position()

    def user_position(self):
        return self.user_buffer.get_position()

    def get_text(self):
        user_text = self.user_buffer.get_matrix()
        reference_text = self.reference_buffer.get_matrix(
            self.valid_user_text_position()
        )
        return user_text + reference_text

    def action(self, screen):
        try:
            action, key = self.listener.listen(screen)
            self.user_buffer.handle_action(action, key)
            self.state.update(action, key)
            self.stats.update(action, self.state.valid(), self.state.running())
        except StoppingSignal as e:
            self.state.signal_stop()
            self.stats.signal_stop()
            if e.silent:
                self.silent_exit = True

    def summarize(self):
        if self.finished:
            self.stats.summarize(self.config.get("summary_template"))
            self.stats.export_to_datafile(self.config.get("summary_datafile"))
            input()

    def exit(self):
        if not self.finished and not self.silent_exit:
            sys.exit(1)


class StoppingSignal(Exception):
    def __init__(self, *args, silent=False, **kwargs):
        super().__init__(*args, **kwargs)
        self.silent = silent



import curses


class Base:
    def paint(self, screen, application):
        pass


class CursorComponent(Base):
    def __init__(self, config):
        super().__init__()
        self.cursor_position = None

    def update(self, screen):
        self.cursor_position = screen.getyx()

    def paint(self, screen, application):
        if self.cursor_position:
            screen.move(*self.cursor_position)


class TextComponent(Base):
    def paint_text(self, screen, text, color):
        screen.addstr(text, curses.color_pair(color))


class StatsComponent(TextComponent):
    def __init__(self, config):
        super().__init__()
        self.color = config.get("stats_color")
        self.template = config.get("stats_template")

    def init(self, screen, application):
        super().init(screen, application)

    def paint(self, screen, application):
        usedy, _ = screen.getyx()
        maxy, _ = screen.getmaxyx()

        text = self.template.format(stats=application.stats)

        texty = len(text.splitlines())

        if texty + usedy + 1 < maxy:
            self.paint_text(screen, text, self.color)


class TopMargin(Base):
    def __init__(self, config):
        super().__init__()
        self.height = config.get("top_margin_percentage") / 100

    def paint(self, screen, application):
        maxy, _ = screen.getmaxyx()
        lines = int(self.height * maxy)
        for line in range(lines):
            screen.addstr("\n")


class TextBox(TextComponent):
    """
    Wraps lines of text elements writing to it nicely
    """

    def __init__(self, config, cursor_component):
        super().__init__()
        self.cursor_component = cursor_component

        self.maxy, self.maxx = None, None
        self.usedy, self.usedx = None, None

        self.left_margin = config.get("left_margin_percentage")

        self.valid_color = config.get("user_input_valid_color")
        self.invalid_color = config.get("user_input_invalid_color")
        self.color = config.get("reference_text_color")

        self.lines_on_screen = config.get("lines_on_screen")

    def clear(self):
        self.maxy, self.maxx = None, None
        self.usedy, self.usedx = None, None

    @property
    def max_line_x(self):
        return int(self.maxx * (100 - self.left_margin - self.left_margin) / 100) - 1

    @property
    def padding(self):
        return " " * int(1 + self.left_margin * self.maxx / 100)

    def lines_to_display(self, application):
        text = application.get_text()
        valid_position = application.valid_user_text_position()
        user_position = application.user_position()

        lines = []
        line = ""
        word = ""

        valid_pointer = (0, 0)
        user_pointer = (0, 0)

        for i, c in enumerate(text):
            word += c

            if len(line + word) > self.max_line_x:
                lines.append(line)
                line = ""

            if not c.isalnum():
                word = str(word)[:-1] + " "
                line += word
                word = ""

        if word:
            line += word

        if line:
            lines.append(line)

        position = 0
        for i, l in enumerate(lines):
            st, end = position, position + len(l)
            if user_position >= st and user_position <= end:
                user_pointer = (i, user_position - st)
            if valid_position >= st and valid_position <= end:
                valid_pointer = (i, valid_position - st)
            position = end

        return lines, valid_pointer, user_pointer

    def paint_line(self, i, line, valid_pointer, user_pointer, screen):
        if i < valid_pointer[0]:
            # easy, we are in written line
            self.paint_text(screen, line, self.valid_color)
            self.cursor_component.update(screen)
            return

        if i > user_pointer[0]:
            # easy, we are in reference line
            self.paint_text(screen, line, self.color)
            return

        valid_text = ""
        invalid_text = ""
        invalid_start = 0

        if i == valid_pointer[0]:
            valid_text = line[: valid_pointer[1]]
            invalid_start = valid_pointer[1]
            invalid_text = line[valid_pointer[1] :]
        if i == user_pointer[0]:
            invalid_text = line[invalid_start : user_pointer[1]]

        reference_text = line[len(invalid_text) + len(valid_text) :]
        invalid_text = invalid_text.replace(" ", "_")

        self.paint_text(screen, valid_text, self.valid_color)
        self.paint_text(screen, invalid_text, self.invalid_color)
        self.cursor_component.update(screen)
        self.paint_text(screen, reference_text, self.color)

    def paint(self, screen, application):
        self.maxy, self.maxx = screen.getmaxyx()
        self.usedy, self.usedx = screen.getyx()

        lines, valid_pointer, user_pointer = self.lines_to_display(application)

        lines_fitting = min((self.maxy - self.usedy, self.lines_on_screen))
        start = 0
        end = len(lines)

        if lines_fitting == 0:
            raise Exception("Too small display!")

        if lines_fitting <= len(lines):
            previous_lines = 1 if lines_fitting > 2 else 0
            next_lines = 1 if lines_fitting > 1 else 0
            start = user_pointer[0] - previous_lines
            end = user_pointer[0] + next_lines
            if start == -1:
                end += 1

        for i, line in enumerate(lines):
            if i >= start and i <= end:
                screen.addstr(self.padding)
                self.paint_line(i, line, valid_pointer, user_pointer, screen)
                screen.addstr("\n")

        self.clear()



from .cli import initialize, get_parser
import sys
import os


def main():
    is_tty = sys.stdin.isatty()
    parser = get_parser()

    if is_tty:
        parser.add_argument("file", metavar="FILE", help="file to type")

    args = parser.parse_args()

    if is_tty:
        with open(os.path.expanduser(args.file)) as f:
            rbuffer = f.read()
    else:
        input_lines = sys.stdin.readlines()

        with open("/dev/tty") as f:
            os.dup2(f.fileno(), 0)

        rbuffer = "".join(input_lines)

    initialize(args.config, rbuffer, args.unclutter_backspace, args.no_cursor)


if __name__ == "__main__":
    main()



from .application import Application
from .interface import Interface
from .components import (
    CursorComponent,
    StatsComponent,
    TextBox,
    TopMargin,
)
from .listener import Listener
from .buffer import UserBuffer, Buffer
from .config import Config
from curses import wrapper
import os
import argparse
import json


def initialize(config_path, rbuffer, backspace_debug, no_cursor):
    try:
        with open(os.path.expanduser(config_path)) as f:
            configmap = json.load(f)
    except FileNotFoundError:
        configmap = {}

    config = Config(configmap)

    reference_buffer = Buffer(rbuffer)
    user_buffer = UserBuffer()

    top_margin = TopMargin(config)
    cursor_component = CursorComponent(config)
    text_box = TextBox(config, cursor_component)
    stats_component = StatsComponent(config)

    listener = Listener(backspace_debug)
    application = Application(listener, user_buffer, reference_buffer, config)

    interface = Interface(
        application,
        [
            top_margin,
            text_box,
            stats_component,
            cursor_component,
        ],
        no_cursor,
    )
    wrapper(interface)

    application.summarize()
    application.exit()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--config",
        "-c",
        metavar="FILE",
        help="configuration file",
        default="~/.config/fasttyper/config.json",
    )
    parser.add_argument(
        "--unclutter-backspace",
        "-b",
        action="store_true",
        help="unclutter backspace, when it raises ctrl+backspace instead",
        default=False,
    )
    parser.add_argument(
        "--no-cursor",
        "-n",
        action="store_true",
        help="disable cursos",
        default=False,
    )
    return parser



from .cli import initialize, get_parser
from random import choice
import os
import requests
import pathlib


MONKEYTYPE_WORDS_PATT = "https://raw.githubusercontent.com/Miodec/monkeytype/master/frontend/static/languages/{language}.json"


def get_words(language):
    home = os.environ.get("HOME", "")
    sfile = os.path.join(home, ".cache", "fasttyper", language)
    words = None

    try:
        with open(sfile) as f:
            words = f.read().splitlines()
    except FileNotFoundError:
        pass

    if words is None:
        source_path = MONKEYTYPE_WORDS_PATT.format(language=language)

        words = requests.get(source_path).json()["words"]

        p = pathlib.Path(os.path.join(home, ".cache", "fasttyper"))
        p.mkdir(parents=True, exist_ok=True)

        with open(sfile, "w") as f:
            f.write("\n".join(words))

    return words


def runner():
    parser = get_parser()

    parser.add_argument(
        "amount", type=int, default=25, help="Amount of words", nargs="?"
    )
    parser.add_argument(
        "language", type=str, default="english", help="Language", nargs="?"
    )

    args = parser.parse_args()
    words = get_words(args.language)

    while True:
        rbuffer = " ".join([choice(words) for _ in range(args.amount)])
        initialize(args.config, rbuffer, args.unclutter_backspace, args.no_cursor)


if __name__ == "__main__":
    runner()



import enum
from .listener import Action
from .application import StoppingSignal


class State(enum.Enum):
    valid = "valid"
    invalid = "invalid"
    finished = "finished"


class StateMachine:
    def __init__(self, application):
        self.state = State.valid
        self.mistake_position = None
        self.application = application

    def running(self):
        return self.state != State.finished

    def signal_stop(self):
        self.state = State.finished

    def signal_start(self):
        self.state = State.valid
        self.mistake_position = None

    def update(self, action, char):
        user_position = self.application.user_buffer.get_position()
        if self.state == State.invalid:
            if action in (Action.del_char, Action.del_word):
                if user_position <= self.mistake_position:
                    self.mistake_position = None
                    self.state = State.valid
        if self.state == State.valid:
            if action in (Action.add_char, Action.add_space, Action.add_newline):
                if char != self.application.reference_buffer.read(user_position - 1):
                    self.mistake_position = user_position - 1
                    self.state = State.invalid
                elif user_position == self.application.reference_buffer.get_lenght():
                    self.state = State.finished
                    self.application.finished = True

    def valid(self):
        return self.state in (State.valid, State.finished)



import enum
from .application import StoppingSignal


class Action(enum.Enum):
    add_char = "add_char"
    add_space = "add_space"
    add_newline = "add_newline"
    del_char = "del_char"
    del_word = "del_word"
    invalid = "invalid"


class Listener:
    def __init__(self, backspace_debug=False):
        self.tabbed = False
        self.backspace_debug = not backspace_debug

    def handle_key(self, key):
        action = Action.invalid

        if key == "\t":
            self.tabbed = True
        elif key == "\n" and self.tabbed:
            raise StoppingSignal(silent=True)
        elif self.backspace_debug and key == chr(8):
            action = Action.del_word
            self.tabbed = False
        elif key == 263 and self.backspace_debug:
            action = Action.del_char
            self.tabbed = False
        elif key == 263:
            action = Action.del_word
            self.tabbed = False
        elif isinstance(key, int):
            self.tabbed = False
        elif key in chr(127):
            action = Action.del_char
            self.tabbed = False
        elif key == " ":
            action = Action.add_space
            self.tabbed = False
        elif key == "\n":
            action = Action.add_newline
            self.tabbed = False
        elif key.isprintable():
            action = Action.add_char
            self.tabbed = False
        else:
            self.tabbed = False

        return action, key

    def listen(self, screen):
        try:
            key = screen.get_wch()
            return self.handle_key(key)
        except KeyboardInterrupt:
            raise StoppingSignal()



from .listener import Action


class Buffer:
    def __init__(self, buffer):
        self.buffer = " ".join(buffer.split())

    def _write(self, data):
        self.buffer += data

    def _del_char(self):
        self.buffer = self.buffer[:-1]

    def _last_char(self):
        if len(self.buffer) > 0:
            return self.buffer[-1]

    def _del_word(self):
        words = self.buffer.split()

        if len(words) > 1:
            self.buffer = " ".join(words[:-1]) + " "
        else:
            self.buffer = ""

    def handle_action(self, action, char):
        if action == Action.add_char:
            self._write(char)
        elif action == Action.add_space:
            self._write(" ")
        elif action == Action.add_newline:
            self._write("\n")
        elif action == Action.del_char:
            self._del_char()
        elif action == Action.del_word:
            self._del_word()

    def get_matrix(self, position=0):
        return self.buffer[position:]

    def get_position(self):
        return len(self.buffer)

    def get_lenght(self):
        return len(self.buffer)

    def read(self, position):
        return self.buffer[position]


class UserBuffer(Buffer):
    def __init__(self):
        super().__init__("")



from setuptools import find_packages, setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="fasttyper",
    version="1.6.0",
    author="Piotr Domanski",
    author_email="pi.domanski@gmail.com",
    description="Minimalistic typing exercise",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ickyicky/fasttyper",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
    ],
    packages=find_packages(),
    python_requires=">=3.6",
    entry_points={
        "console_scripts": [
            "fasttyper=fasttyper.runner:runner",
        ]
    },
)
